import os
from argparse import HelpFormatter
from multiprocessing import cpu_count

try:
    from gooey import GooeyParser as ArgumentParser
except ImportError:
    from argparse import ArgumentParser
    HAS_GOOEY = False
else:
    HAS_GOOEY = True

from deeplc import __version__


def parse_arguments(gui=False):
    """Read arguments from the CLI or GUI."""
    if gui and not HAS_GOOEY:
        raise ImportError(
            "Missing dependency `gooey` required to start graphical user interface. "
            "Install `gooey` or use the command line interface."
        )

    # Define Gooey-specific arguments upfront to be enabled only if Gooey is available
    gooey_args = {
        "io_args": {"gooey_options": {'columns':2}},
        "file_pred": {
            "widget": "FileChooser",
            "metavar": "Input peptides for prediction (required)"
        },
        "file_cal": {"widget": "FileChooser"},
        "file_pred_out": {"widget": "FileSaver"},
        "plot_predictions": {
            "widget": "BlockCheckbox",
            "gooey_options": {"checkbox_label": "Enable"},
            "metavar": "Plot predictions",
        },
        "model_cal_args": {"gooey_options": {'columns':2}},
        "file_model": {"widget": "MultiFileChooser"},
        "pygam_calibration": {
            "widget": "BlockCheckbox",
            "gooey_options": {"checkbox_label": "Use pyGAM calibration"},
            "metavar": "Use pyGAM calibration"
        },
        "transfer_learning": {
            "widget": "BlockCheckbox",
            "gooey_options": {"checkbox_label": "Use transfer learning calibration"},
            "metavar": "Use transfer learning calibration"
        },
        "split_cal": {"gooey_options": {"visible": False}},
        "dict_divider": {"gooey_options": {"visible": False}},
        "advanced_args": {"gooey_options": {'columns':2}},
        "use_library": {"widget": "FileChooser"},
        "write_library": {
            "widget": "BlockCheckbox",
            "gooey_options": {"checkbox_label": "Enable"},
            "metavar": "Append to prediction library",
        },
        "batch_num": {
            "widget": "IntegerField",
            "gooey_options": {"min": 10e3, "max": 10e9, "increment": 1000},
        },
        "n_threads": {
            "widget": "Slider",
            "gooey_options": {"min": 1, "max": cpu_count(), "increment": 1},
        },
        "log_level": {"widget": "Dropdown"},
        "version": {"gooey_options": {"visible": False}},
        "help": {"gooey_options": {"visible": False}},
    }
    if not gui:
        gooey_args = {k: {} for k, v in gooey_args.items()}

    parser = ArgumentParser(
        prog="DeepLC",
        description=(
            "Retention time prediction for (modified) peptides using deep "
            "learning."),
        usage="deeplc [OPTIONS] --file_pred <peptide_file>",
        formatter_class=lambda prog: HelpFormatter(prog, max_help_position=42),
        add_help=False,
    )

    io_args = parser.add_argument_group(
        "Input and output files",
        **gooey_args["io_args"]
    )
    io_args.add_argument(
        "--file_pred",
        required=True,
        type=str,
        help="path to peptide CSV file for which to make predictions (required)",
        **gooey_args["file_pred"]
    )
    io_args.add_argument(
        "--file_cal",
        type=str,
        default=None,
        metavar="Input peptides for calibration" if gui else "",
        help=(
            "path to peptide CSV file with retention times to use for "
            "calibration"
        ),
        **gooey_args["file_cal"]
    )
    io_args.add_argument(
        "--file_pred_out",
        type=str,
        default=None,
        metavar="Output filename" if gui else "",
        help="path to write output file with predictions",
        **gooey_args["file_pred_out"]
    )

    model_cal_args = parser.add_argument_group(
        "Model and calibration",
        **gooey_args["model_cal_args"]
    )
    model_cal_args.add_argument(
        "--file_model",
        nargs="+",
        default=None,
        metavar="Model file(s)" if gui else "",
        help=(
            "path to prediction model(s); leave empty to select the best of "
            "the default models based on the calibration peptides"
        ),
        **gooey_args["file_model"],
    )

    #calibration_group = model_cal_args.add_mutually_exclusive_group(
    #    **gooey_args["calibration_group"]
    #)
    #model_cal_args.add_argument(
    #    "--pygam_calibration",
    #    dest="pygam_calibration",
    #    action="store_true",
    #    #default=True,
    #    help=(
    #        "use pyGAM generalized additive model as calibration method; "
    #        "recommended; default"
    #    ),
    #    **gooey_args["pygam_calibration"]
    #)
    
    model_cal_args.add_argument(
        "--transfer_learning",
        dest="transfer_learning",
        action="store_true",
        default=False,
        help="use transfer learning as calibration method",
        **gooey_args["transfer_learning"]
    )

    """
    calibration_group.add_argument(
        "--legacy_calibration",
        dest="pygam_calibration",
        action="store_false",
        help="use legacy simple piecewise linear fit as calibration method",
        **gooey_args["legacy_calibration"]
    )
    """

    model_cal_args.add_argument(
        "--split_cal",
        type=int,
        dest="split_cal",
        default=50,
        metavar="split cal" if gui else "",
        help=(
            "number of splits in the chromatogram for piecewise linear "
            "calibration fit"
        ),
        **gooey_args["split_cal"]
    )
    model_cal_args.add_argument(
        "--dict_divider",
        type=int,
        dest="dict_divider",
        default=50,
        metavar="dict divider" if gui else "",
        help=(
            "sets precision for fast-lookup of retention times for "
            "calibration; e.g. 10 means a precision of 0.1 between the "
            "calibration anchor points"
        ),
        **gooey_args["dict_divider"]
    )

    advanced_args = parser.add_argument_group(
        "Advanced configuration", **gooey_args["advanced_args"]
    )
    """
    advanced_args.add_argument(
        "--use_library",
        dest="use_library",
        action='store',
        default=False,
        metavar="Select prediction library file" if gui else "",
        help=(
            "library file with previous predictions for faster results to "
            "read from, or to write to"
        ),
        **gooey_args["use_library"]
    )
    """

    """
    advanced_args.add_argument(
        "--write_library",
        dest="write_library",
        action='store_true',
        default=False,
        help="append new predictions to library for faster future results",
        **gooey_args["write_library"]
    )
    """

    advanced_args.add_argument(
        "--batch_num",
        type=int,
        dest="batch_num",
        default=250000,
        metavar="Batch size" if gui else "",
        help=(
            "prediction batch size (in peptides); lower to decrease memory "
            "footprint; default=250000"
        ),
        **gooey_args["batch_num"]
    )
    advanced_args.add_argument(
        "--n_threads",
        type=int,
        default=max(cpu_count(), 16),
        metavar="Parallelization " if gui else "",
        help="number of CPU threads to use; default=all with max of 16",
        **gooey_args["n_threads"]
    )
    advanced_args.add_argument(
        "--log_level",
        type=str,
        action="store",
        dest="log_level",
        default='info',
        metavar="Logging level" if gui else "",
        choices=["debug","info","warning","error","critical"],
        help="verbosity of logging messages; default=info",
        **gooey_args["log_level"]
    )
    advanced_args.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        **gooey_args["version"]
    )
    advanced_args.add_argument(
        "-h",
        "--help",
        action="help",
        help="show this help message and exit",
        **gooey_args["help"]
    )

    results = parser.parse_args()

    if not results.file_pred_out:
        results.file_pred_out = os.path.splitext(results.file_pred)[0] + '_deeplc_predictions.csv'

    return results



from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def scatter(
    df: pd.DataFrame,
    predicted_column: str = "Predicted retention time",
    observed_column: str = "Observed retention time",
) -> go.Figure:
    """
    Plot a scatter plot of the predicted vs. observed retention times.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing the predicted and observed retention times.
    predicted_column : str, optional
        Name of the column containing the predicted retention times, by default
        ``Predicted retention time``.
    observed_column : str, optional
        Name of the column containing the observed retention times, by default
        ``Observed retention time``.

    """
    # Draw scatter
    fig = px.scatter(
        df,
        x=observed_column,
        y=predicted_column,
        opacity=0.3,
    )
    
    # Draw diagonal line
    fig.add_scatter(
        x=[min(df[observed_column]), max(df[observed_column])],
        y=[min(df[observed_column]), max(df[observed_column])],
        mode='lines',
        line=dict(color='red', width=3, dash='dash'),
    )
    
    # Hide legend
    fig.update_layout(
        title="Predicted vs. observed retention times",
        showlegend=False,
        xaxis_title="Observed retention time",
        yaxis_title="Predicted retention time",
    )
    
    return fig


def distribution_baseline(
    df: pd.DataFrame,
    predicted_column: str = "Predicted retention time",
    observed_column: str = "Observed retention time",
) -> go.Figure:
    """
    Plot a distribution plot of the relative mean absolute error of the current
    DeepLC performance compared to the baseline performance.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing the predicted and observed retention times.
    predicted_column : str, optional
        Name of the column containing the predicted retention times, by default
        ``Predicted retention time``.
    observed_column : str, optional
        Name of the column containing the observed retention times, by default
        ``Observed retention time``.

    """
    # Get baseline data
    baseline_df = pd.read_csv(
        Path(__file__)
        .absolute()
        .parent.joinpath("baseline_performance/baseline_predictions.csv")
    )
    baseline_df["rel_mae_best"] = baseline_df[
        ["rel_mae_transfer_learning", "rel_mae_new_model", "rel_mae_calibrate"]
    ].min(axis=1)
    baseline_df.fillna(0.0, inplace=True)

    # Calculate current RMAE and percentile compared to baseline
    mae = sum(abs(df[observed_column] - df[predicted_column])) / len(df.index)
    mae_rel = (mae / max(df[observed_column])) * 100
    percentile = round(
        (baseline_df["rel_mae_transfer_learning"] < mae_rel).mean() * 100, 1
    )

    # Calculate x-axis range with 5% padding
    all_values = np.append(baseline_df["rel_mae_transfer_learning"].values, mae_rel)
    padding = (all_values.max() - all_values.min()) / 20  # 5% padding
    x_min = all_values.min() - padding
    x_max = all_values.max() + padding

    # Make labels human-readable
    hover_label_mapping = {
        "train_number": "Training dataset size",
        "rel_mae_transfer_learning": "RMAE with transfer learning",
        "rel_mae_new_model": "RMAE with new model from scratch",
        "rel_mae_calibrate": "RMAE with calibrating existing model",
        "rel_mae_best": "RMAE with best method",
    }
    label_mapping = hover_label_mapping.copy()
    label_mapping.update({"Unnamed: 0": "Dataset"})

    # Generate plot
    fig = px.histogram(
        data_frame=baseline_df,
        x="rel_mae_best",
        marginal="rug",
        hover_data=hover_label_mapping.keys(),
        hover_name="Unnamed: 0",
        labels=label_mapping,
        opacity=0.8,
    )
    fig.add_vline(
        x=mae_rel,
        line_width=3,
        line_dash="dash",
        line_color="red",
        annotation_text=f"Current performance (percentile {percentile}%)",
        annotation_position="top left",
        name="Current performance",
        row=1,
    )
    fig.update_xaxes(range=[x_min, x_max])
    fig.update_layout(
        title=(
            f"Current DeepLC performance compared to {len(baseline_df.index)} "
            "datasets"
        ),
        xaxis_title="Relative mean absolute error (%)",
    )

    return fig



"""
Main code used to generate LC retention time predictions.

This provides the main interface. For the library versions see the .yml file
"""


__author__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__license__ = "Apache License, Version 2.0"
__maintainer__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__email__ = ["Robbin.Bouwmeester@ugent.be", "Ralf.Gabriels@ugent.be"]
__credits__ = [
    "Robbin Bouwmeester",
    "Ralf Gabriels",
    "Arthur Declercq"
    "Lennart Martens",
    "Sven Degroeve",
]


# Default models, will be used if no other is specified. If no best model is
# selected during calibration, the first model in the list will be used.
import os

deeplc_dir = os.path.dirname(os.path.realpath(__file__))
DEFAULT_MODELS = [
    "mods/full_hc_PXD005573_mcp_1fd8363d9af9dcad3be7553c39396960.hdf5",
    "mods/full_hc_PXD005573_mcp_8c22d89667368f2f02ad996469ba157e.hdf5",
    "mods/full_hc_PXD005573_mcp_cb975cfdd4105f97efa0b3afffe075cc.hdf5"
]
DEFAULT_MODELS = [os.path.join(deeplc_dir, dm) for dm in DEFAULT_MODELS]

LIBRARY = {}

import os
import sys
import copy
import gc
import logging
import multiprocessing
import multiprocessing.dummy
import pickle
import warnings
from configparser import ConfigParser
from tempfile import TemporaryDirectory
from copy import deepcopy
import random
import math

# If CLI/GUI/frozen: disable Tensorflow info and warnings before importing
IS_CLI_GUI = os.path.basename(sys.argv[0]) in ["deeplc", "deeplc-gui"]
IS_FROZEN = getattr(sys, "frozen", False)
if IS_CLI_GUI or IS_FROZEN:
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    logging.getLogger("tensorflow").setLevel(logging.CRITICAL)
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    warnings.filterwarnings("ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", category=UserWarning)

# Supress warnings (or at least try...)
logging.getLogger('tensorflow').setLevel(logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.eager import context
from tensorflow.keras.models import load_model
import h5py

from deeplc._exceptions import CalibrationError, DeepLCError
from deeplc.trainl3 import train_en

from psm_utils.io.peptide_record import peprec_to_proforma
from psm_utils.psm import PSM
from psm_utils.psm_list import PSMList
from psm_utils.io import read_file
from psm_utils.io import write_file

from deeplcretrainer import deeplcretrainer

# "Custom" activation function
lrelu = lambda x: tf.keras.activations.relu(x, alpha=0.1, max_value=20.0)


try:
    from tensorflow.compat.v1.keras.backend import set_session
except ImportError:
    from tensorflow.keras.backend import set_session
try:
    from tensorflow.compat.v1.keras.backend import clear_session
except ImportError:
    from tensorflow.keras.backend import clear_session
try:
    from tensorflow.compat.v1.keras.backend import get_session
except ImportError:
    from tensorflow.keras.backend import get_session

# Set to force CPU calculations
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"


# Set for TF V1.0 (counters some memory problems of nvidia 20 series GPUs)
# config = tf.ConfigProto()
# config.gpu_options.allow_growth = True
# session = tf.Session(config=config)

# Set for TF V2.0 (counters some memory problems of nvidia 20 series GPUs)
# config = tf.compat.v1.ConfigProto()
# config.gpu_options.allow_growth = True
# session = tf.compat.v1.Session(config=config)

# Feature extraction
from deeplc.feat_extractor import FeatExtractor
from pygam import LinearGAM, s

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=FutureWarning)

logger = logging.getLogger(__name__)

def split_list(a, n):
    k, m = divmod(len(a), n)
    return (a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))

def divide_chunks(l, n):
    for i in range(0, len(l), n): 
        yield l[i:i + n]

def reset_keras():
    """Reset Keras session."""
    sess = get_session()
    clear_session()
    sess.close()
    gc.collect()
    # Set to force CPU calculations
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"


class DeepLC:
    """
    DeepLC predictor.

    Parameters
    ----------
    main_path : str
        main path of module
    path_model : str, optional
        path to prediction model(s); leave empty to select the best of the
        default models based on the calibration peptides
    verbose : bool, default=True
        turn logging on/off
    bin_dist : float, default=2
        TODO
    dict_cal_divider : int, default=50
        sets precision for fast-lookup of retention times for calibration; e.g.
        10 means a precision of 0.1 between the calibration anchor points
    split_cal : int, default=50
        number of splits in the chromatogram for piecewise linear calibration
        fit
    n_jobs : int, optional
        number of CPU threads to use
    config_file : str, optional
        path to configuration file
    f_extractor : object :: deeplc.FeatExtractor, optional
        deeplc.FeatExtractor object to use
    cnn_model : bool, default=True
        use CNN model or not
    batch_num : int, default=250000
        prediction batch size (in peptides); lower to decrease memory footprint
    write_library : bool, default=False
        append new predictions to library for faster future results; requires
        `use_library` option
    use_library : str, optional
        library file with previous predictions for faster results to read from,
        or to write to
    reload_library : bool, default=False
        reload prediction library

    Methods
    -------
    calibrate_preds(seqs=[], mods=[], identifiers=[], measured_tr=[], correction_factor=1.0, seq_df=None, use_median=True)
        Find best model and calibrate
    make_preds(seqs=[], mods=[], identifiers=[], calibrate=True, seq_df=None, correction_factor=1.0, mod_name=None)
        Make predictions

    """

    library = {}

    # TODO have a CCS flag here
    def __init__(
        self,
        main_path=os.path.dirname(os.path.realpath(__file__)),
        path_model=None,
        verbose=True,
        bin_dist=2,
        dict_cal_divider=50,
        split_cal=50,
        n_jobs=None,
        config_file=None,
        f_extractor=None,
        cnn_model=True,
        batch_num=250000,
        batch_num_tf=1024,
        write_library=False,
        use_library=None,
        reload_library=False,
        pygam_calibration=True,
        deepcallc_mod=False,
        deeplc_retrain=False,
        predict_ccs=False,
        n_epochs=20,
    ):  
        # if a config file is defined overwrite standard parameters
        if config_file:
            cparser = ConfigParser()
            cparser.read(config_file)
            dict_cal_divider = cparser.getint("DeepLC", "dict_cal_divider")
            split_cal = cparser.getint("DeepLC", "split_cal")
            n_jobs = cparser.getint("DeepLC", "n_jobs")

        self.main_path = main_path
        self.verbose = verbose
        self.bin_dist = bin_dist
        self.calibrate_dict = {}
        self.calibrate_min = float("inf")
        self.calibrate_max = 0
        self.n_epochs = n_epochs
        self.cnn_model = cnn_model

        self.batch_num = batch_num
        self.batch_num_tf = batch_num_tf
        self.dict_cal_divider = dict_cal_divider
        self.split_cal = split_cal
        self.n_jobs = n_jobs

        max_threads = multiprocessing.cpu_count()
        if not self.n_jobs:
            self.n_jobs = max_threads
        elif self.n_jobs > max_threads:
            self.n_jobs = max_threads

        self.use_library = use_library
        self.write_library = write_library

        self.reload_library = reload_library

        try:
            tf.config.threading.set_intra_op_parallelism_threads(n_jobs)
        except RuntimeError:
            logger.warning(
                "DeepLC tried to set intra op threads, but was unable to do so."
            )

        if "NUMEXPR_MAX_THREADS" not in os.environ:
            os.environ["NUMEXPR_MAX_THREADS"] = str(n_jobs)

        if path_model:
            self.model = path_model
        else:
            self.model = DEFAULT_MODELS

        if f_extractor:
            self.f_extractor = f_extractor
        else:
            self.f_extractor = FeatExtractor()

        self.pygam_calibration = pygam_calibration
        self.deeplc_retrain = deeplc_retrain

        if self.pygam_calibration:
            from pygam import LinearGAM, s

        self.deepcallc_mod = deepcallc_mod

        self.predict_ccs = predict_ccs

        if self.deepcallc_mod:
            self.write_library = False
            self.use_library = None
            self.reload_library = False

    def __str__(self):
        return """
  _____                  _      _____
 |  __ \                | |    / ____|
 | |  | | ___  ___ _ __ | |   | |
 | |  | |/ _ \/ _ \ '_ \| |   | |
 | |__| |  __/  __/ |_) | |___| |____
 |_____/ \___|\___| .__/|______\_____|
                  | |
                  |_|
              """

    def do_f_extraction(self, seqs, mods, identifiers, charges=[]):
        """
        Extract all features we can extract; without parallelization; use if you
        want to run feature extraction with a single core

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : list
            identifiers of the peptides; should correspond to seqs and mods

        Returns
        -------
        pd.DataFrame
            feature matrix
        """
        list_of_psms = []

        if not self.predict_ccs:
            for seq,mod,ident in zip(seqs,mods,identifiers):
                list_of_psms.append(PSM(peptide=peprec_to_proforma(seq,mod),spectrum_id=ident))
        else:
            for seq,mod,ident,z in zip(seqs,mods,identifiers,charges):
                list_of_psms.append(PSM(peptide=peprec_to_proforma(seq,mod,z),spectrum_id=ident))

        psm_list = PSMList(psm_list=list_of_psms)

        return self.f_extractor.full_feat_extract(psm_list)

    def do_f_extraction_pd(self,
                           df_instances,
                           charges=[]):
        """
        Extract all features we can extract; without parallelization; use if
        you want to run feature extraction with a single thread; and use a
        defined dataframe

        Parameters
        ----------
        df_instances : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index)

        Returns
        -------
        pd.DataFrame
            feature matrix
        """

        list_of_psms = []
        if len(charges) == 0:
            for seq,mod,ident in zip(df_instances["seq"],df_instances["modifications"],df_instances.index):
                list_of_psms.append(PSM(peptide=peprec_to_proforma(seq,mod),spectrum_id=ident))
        else:
            for seq,mod,ident,z in zip(df_instances["seq"],df_instances["modifications"],df_instances.index,charges=df_instances["charges"]):
                list_of_psms.append(PSM(peptide=peprec_to_proforma(seq,mod,charge=z),spectrum_id=ident))

        psm_list = PSMList(psm_list=list_of_psms)

        return self.f_extractor.full_feat_extract(psm_list)

    def do_f_extraction_pd_parallel(self, df_instances):
        """
        Extract all features we can extract; with parallelization; use if you
        want to run feature extraction with multiple threads; and use a defined
        dataframe

        Parameters
        ----------
        df_instances : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index)

        Returns
        -------
        pd.DataFrame
            feature matrix
        """
        #self.n_jobs = 1

        df_instances_split = np.array_split(df_instances, math.ceil(self.n_jobs/4.0))
        if multiprocessing.current_process().daemon:
            logger.warning(
                "DeepLC is running in a daemon process. Disabling multiprocessing as daemonic processes can't have children."
            )
            pool = multiprocessing.dummy.Pool(1)
        else:
            pool = multiprocessing.Pool(math.ceil(self.n_jobs/4.0))

        if self.n_jobs == 1:
            df = self.do_f_extraction_pd(df_instances)
        else:
            df = pd.concat(
                pool.map(
                    self.do_f_extraction_pd,
                    df_instances_split))
            pool.close()
            pool.join()
        return df

    def do_f_extraction_psm_list(
                        self,
                        psm_list
            ):
        """
        Extract all features we can extract; without parallelization; use if
        you want to run feature extraction with a single thread; and use a
        defined dataframe

        Parameters
        ----------
        df_instances : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index)

        Returns
        -------
        pd.DataFrame
            feature matrix
        """
        return self.f_extractor.full_feat_extract(psm_list)

    def do_f_extraction_psm_list_parallel(
                        self,
                        psm_list
            ):
        """
        Extract all features we can extract; without parallelization; use if
        you want to run feature extraction with a single thread; and use a
        defined dataframe

        Parameters
        ----------
        df_instances : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index)

        Returns
        -------
        pd.DataFrame
            feature matrix
        """
        # TODO for multiproc I am still expecting a pd dataframe, this is not the case anymore, they are dicts
        self.n_jobs = 1
        logger.debug("prepare feature extraction")
        if multiprocessing.current_process().daemon:
            logger.warning("DeepLC is running in a daemon process. Disabling multiprocessing as daemonic processes can't have children.")
            psm_list_split = split_list(psm_list, self.n_jobs)
            pool = multiprocessing.dummy.Pool(1)
        elif self.n_jobs > 1:
            psm_list_split = split_list(psm_list, self.n_jobs)
            pool = multiprocessing.Pool(self.n_jobs)

        if self.n_jobs == 1:
            logger.debug("start feature extraction")
            all_feats = self.do_f_extraction_psm_list(psm_list)
            logger.debug("got feature extraction results")
        else:
            logger.debug("start feature extraction")
            all_feats_async = pool.map_async(
                    self.do_f_extraction_psm_list,
                    psm_list_split)

            logger.debug("wait for feature extraction")
            all_feats_async.wait()
            logger.debug("get feature extraction results")
            all_feats = pd.concat(all_feats_async.get())
            logger.debug("got feature extraction results")

            pool.close()
            pool.join()

        return all_feats

    def calibration_core(self,uncal_preds,cal_dict,cal_min,cal_max):
        cal_preds = []
        if len(uncal_preds) == 0:
            return np.array(cal_preds)
        if self.pygam_calibration:
            cal_preds = cal_dict.predict(uncal_preds)
        else:
            for uncal_pred in uncal_preds:
                try:
                    slope, intercept = cal_dict[str(round(uncal_pred, self.bin_dist))]
                    cal_preds.append(slope * (uncal_pred) + intercept)
                except KeyError:
                    # outside of the prediction range ... use the last
                    # calibration curve
                    if uncal_pred <= cal_min:
                        slope, intercept = cal_dict[str(round(cal_min, self.bin_dist))]
                        cal_preds.append(slope * (uncal_pred) + intercept)
                    elif uncal_pred >= cal_max:
                        slope, intercept = cal_dict[str(round(cal_max, self.bin_dist))]
                        cal_preds.append(slope * (uncal_pred) + intercept)
                    else:
                        slope, intercept = cal_dict[str(round(cal_max, self.bin_dist))]
                        cal_preds.append(slope * (uncal_pred) + intercept)
        return np.array(cal_preds)

    def make_preds_core_library(self,
                                psm_list=[],
                                calibrate=True,
                                mod_name=None
                                ):
        ret_preds = []
        for psm in psm_list:
            ret_preds.append(LIBRARY[psm.peptidoform.proforma+"|"+mod_name])

        if calibrate:
            try:
                ret_preds = self.calibration_core(ret_preds,self.calibrate_dict[mod_name],self.calibrate_min[mod_name],self.calibrate_max[mod_name])
            except:
                ret_preds = self.calibration_core(ret_preds,self.calibrate_dict,self.calibrate_min,self.calibrate_max)
        
        return ret_preds

    def make_preds_core(self,
                        X=[], 
                        X_sum=[], 
                        X_global=[], 
                        X_hc=[],
                        psm_list=[],
                        calibrate=True,
                        mod_name=None
                        ):
        """
        Make predictions for sequences
        Parameters
        ----------
        seq_df : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index); will use parallel
            by default!
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : list
            identifiers of the peptides; should correspond to seqs and mods
        calibrate : boolean
            calibrate predictions or just return the predictions
        correction_factor : float
            correction factor to apply to predictions
        mod_name : str or None
            specify a model to use instead of the model assigned originally to
            this instance of the object
        Returns
        -------
        np.array
            predictions
        """
        if calibrate:
            assert (
                self.calibrate_dict
            ), "DeepLC instance is not yet calibrated.\
                                        Calibrate before making predictions, or use calibrate=False"

        if len(X) == 0 and len(psm_list) > 0:
            if self.verbose:
                logger.debug("Extracting features for the CNN model ...")
            #X = self.do_f_extraction_psm_list(psm_list)
            X = self.do_f_extraction_psm_list_parallel(psm_list)

            X_sum = np.stack(list(X["matrix_sum"].values()))
            X_global = np.concatenate((np.stack(list(X["matrix_all"].values())),
                                    np.stack(list(X["pos_matrix"].values()))),
                                    axis=1)
            X_hc = np.stack(list(X["matrix_hc"].values()))
            X = np.stack(list(X["matrix"].values()))
        elif len(X) == 0 and len(psm_list) == 0:
            return []

        ret_preds = []

        mod = load_model(
                    mod_name,
                    custom_objects={'<lambda>': lrelu}
                )
        try:
            X
            ret_preds = mod.predict(
                [X, X_sum, X_global, X_hc], batch_size=self.batch_num_tf).flatten()
        except UnboundLocalError:
            logger.debug("X is empty, skipping...")
            ret_preds = []

        if calibrate:
            try:
                ret_preds = self.calibration_core(ret_preds,self.calibrate_dict[mod_name],self.calibrate_min[mod_name],self.calibrate_max[mod_name])
            except:
                ret_preds = self.calibration_core(ret_preds,self.calibrate_dict,self.calibrate_min,self.calibrate_max)
        
        clear_session()
        gc.collect()
        return ret_preds

    def make_preds(self,
                   psm_list=None,
                   infile="",
                   calibrate=True,
                   seq_df=None,
                   mod_name=None):
        """
        Make predictions for sequences, in batches if required.

        Parameters
        ----------
        seq_df : object :: pd.DataFrame
            dataframe containing the sequences (column:seq), modifications
            (column:modifications) and naming (column:index); will use parallel
            by default!
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : list
            identifiers of the peptides; should correspond to seqs and mods
        calibrate : boolean
            calibrate predictions or just return the predictions
        correction_factor : float
            correction factor to apply to predictions
        mod_name : str or None
            specify a model to use instead of the model assigned originally to
            this instance of the object

        Returns
        -------
        np.array
            predictions
        """
        if type(seq_df) == pd.core.frame.DataFrame:
            list_of_psms = []
            if self.predict_ccs:
                for seq,mod,ident,z in zip(seq_df["seq"],seq_df["modifications"],seq_df.index,seq_df["charge"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod,charge=z),spectrum_id=ident))
            else:
                for seq,mod,ident in zip(seq_df["seq"],seq_df["modifications"],seq_df.index):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident))
            psm_list = PSMList(psm_list=list_of_psms)
        
        if len(infile) > 0:
            psm_list = read_file(infile)
            if "msms" in infile and ".txt" in infile:
                mapper = pd.read_csv(os.path.join(os.path.dirname(os.path.realpath(__file__)), "unimod/map_mq_file.csv"),index_col=0)["value"].to_dict()
                psm_list.rename_modifications(mapper)

        ret_preds_batches = []
        for psm_list_t in divide_chunks(psm_list, self.batch_num):
            ret_preds = []
            if len(psm_list_t) > 0:
                if self.verbose:
                    logger.debug("Extracting features for the CNN model ...")

                X = self.do_f_extraction_psm_list_parallel(psm_list_t)
                X_sum = np.stack(list(X["matrix_sum"].values()))
                X_global = np.concatenate((np.stack(list(X["matrix_all"].values())),
                                        np.stack(list(X["pos_matrix"].values()))),
                                        axis=1)
                X_hc = np.stack(list(X["matrix_hc"].values()))
                X = np.stack(list(X["matrix"].values()))
            else:
                return []

            if isinstance(self.model, dict):
                for m_group_name,m_name in self.model.items():
                    ret_preds.append(self.make_preds_core(X=X, 
                                        X_sum=X_sum, 
                                        X_global=X_global, 
                                        X_hc=X_hc,
                                        calibrate=calibrate,
                                        mod_name=m_name))
                ret_preds = np.array([sum(a)/len(a) for a in zip(*ret_preds)])
            elif mod_name is not None:
                ret_preds = self.make_preds_core(X=X, 
                                                X_sum=X_sum, 
                                                X_global=X_global, 
                                                X_hc=X_hc,
                                                calibrate=calibrate,
                                                mod_name=mod_name)
            elif isinstance(self.model, list):
                for m_name in self.model:
                    ret_preds.append(self.make_preds_core(X=X, 
                                        X_sum=X_sum, 
                                        X_global=X_global, 
                                        X_hc=X_hc,
                                        calibrate=calibrate,
                                        mod_name=m_name))
                ret_preds = np.array([sum(a)/len(a) for a in zip(*ret_preds)])
            else:
                ret_preds = self.make_preds_core(X=X, 
                                                X_sum=X_sum, 
                                                X_global=X_global, 
                                                X_hc=X_hc,
                                                calibrate=calibrate,
                                                mod_name=self.model)
            ret_preds_batches.extend(ret_preds)

        return ret_preds_batches
        # TODO make this multithreaded
        # should be possible with the batched list

    def calibrate_preds_func_pygam(self,
                                   psm_list=None,
                                   correction_factor=1.0,
                                   seq_df=None,
                                   measured_tr=None,
                                   use_median=True,
                                   mod_name=None):
        # TODO make a df to psm_list function
        # TODO make sure either psm_list or seq_df is supplied
        if type(seq_df) == pd.core.frame.DataFrame:
            list_of_psms = []
            # TODO include charge here
            if self.predict_ccs:
                for seq,mod,ident,tr,z in zip(seq_df["seq"],seq_df["modifications"],seq_df.index,seq_df["tr"],seq_df["charge"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod,charge=z),spectrum_id=ident,retention_time=tr))
            else:
                for seq,mod,ident,tr in zip(seq_df["seq"],seq_df["modifications"],seq_df.index,seq_df["tr"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident,retention_time=tr))
            psm_list = PSMList(psm_list=list_of_psms)

            measured_tr = [psm.retention_time for psm in psm_list]

        predicted_tr = self.make_preds(
            psm_list,
            calibrate=False,
            mod_name=mod_name)

        # sort two lists, predicted and observed based on measured tr
        tr_sort = [
            (mtr, ptr)
            for mtr, ptr in sorted(
                zip(measured_tr, predicted_tr), key=lambda pair: pair[1]
            )
        ]
        measured_tr = np.array([mtr for mtr, ptr in tr_sort], dtype=np.float32)
        predicted_tr = np.array([ptr for mtr, ptr in tr_sort], dtype=np.float32)

        # predicted_tr = list(predicted_tr)
        # measured_tr = list(measured_tr)

        gam_model_cv = LinearGAM(s(0), verbose=True).fit(predicted_tr, measured_tr)
        calibrate_min = min(predicted_tr)
        calibrate_max = max(predicted_tr)
        return calibrate_min, calibrate_max, gam_model_cv

    def calibrate_preds_func(self,
                             psm_list=None,
                             correction_factor=1.0,
                             seq_df=None,
                             use_median=True,
                             mod_name=None):
        """
        Make calibration curve for predictions

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : list
            identifiers of the peptides; should correspond to seqs and mods
        measured_tr : list
            measured tr of the peptides; should correspond to seqs, identifiers,
            and mods
        correction_factor : float
            correction factor that needs to be applied to the supplied measured
            trs
        seq_df : object :: pd.DataFrame
            a pd.DataFrame that contains the sequences, modifications and
            observed retention times to fit a calibration curve
        use_median : boolean
            flag to indicate we need to use the median valuein a window to
            perform calibration
        mod_name
            specify a model to use instead of the model assigned originally to
            this instance of the object

        Returns
        -------
        float
            the minimum value where a calibration curve was fitted, lower values
            will be extrapolated from the minimum fit of the calibration curve
        float
            the maximum value where a calibration curve was fitted, higher values
            will be extrapolated from the maximum fit of the calibration curve
        dict
            dictionary with keys for rounded tr, and the values concern a linear
            model that should be applied to do calibration (!!! what is the
            shape of this?)
        """
        if type(seq_df) == pd.core.frame.DataFrame:
            list_of_psms = []
            # TODO include charge here
            if self.predict_ccs:
                for seq,mod,tr,ident,z in zip(seq_df["seq"],seq_df["modifications"],seq_df["tr"],seq_df.index,seq_df["charge"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod,charge=z),spectrum_id=ident,retention_time=tr))
            else:
                for seq,mod,tr,ident in zip(seq_df["seq"],seq_df["modifications"],seq_df["tr"],seq_df.index):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident,retention_time=tr))
            psm_list = PSMList(psm_list=list_of_psms)
        
        measured_tr = [psm.retention_time for psm in psm_list]

        predicted_tr = self.make_preds(
            psm_list,
            calibrate=False,
            mod_name=mod_name)

        # sort two lists, predicted and observed based on measured tr
        tr_sort = [
            (mtr, ptr)
            for mtr, ptr in sorted(
                zip(measured_tr, predicted_tr), key=lambda pair: pair[1]
            )
        ]
        measured_tr = np.array([mtr for mtr, ptr in tr_sort])
        predicted_tr = np.array([ptr for mtr, ptr in tr_sort])

        mtr_mean = []
        ptr_mean = []

        calibrate_dict = {}
        calibrate_min = float("inf")
        calibrate_max = 0

        if self.verbose:
            logger.debug(
                "Selecting the data points for calibration (used to fit the linear models between)"
            )
        # smooth between observed and predicted
        split_val = predicted_tr[-1] / self.split_cal

        for range_calib_number in np.arange(0.0, predicted_tr[-1], split_val):
            ptr_index_start = np.argmax(predicted_tr >= range_calib_number)
            ptr_index_end = np.argmax(predicted_tr >= range_calib_number + split_val)

            # no points so no cigar... use previous points
            if ptr_index_start >= ptr_index_end:
                logger.debug(
                    "Skipping calibration step, due to no points in the "
                    "predicted range (are you sure about the split size?): "
                    "%s,%s",
                    range_calib_number,
                    range_calib_number + split_val,
                )
                continue

            mtr = measured_tr[ptr_index_start:ptr_index_end]
            ptr = predicted_tr[ptr_index_start:ptr_index_end]

            if use_median:
                mtr_mean.append(np.median(mtr))
                ptr_mean.append(np.median(ptr))
            else:
                mtr_mean.append(sum(mtr) / len(mtr))
                ptr_mean.append(sum(ptr) / len(ptr))

        if self.verbose:
            logger.debug("Fitting the linear models between the points")

        if self.split_cal >= len(measured_tr):
            raise CalibrationError(
                "Not enough measured tr ({}) for the chosen number of splits ({}). "
                "Choose a smaller split_cal parameter or provide more peptides for "
                "fitting the calibration curve.".format(
                    len(measured_tr), self.split_cal
                )
            )
        if len(mtr_mean) == 0:
            raise CalibrationError(
                "The measured tr list is empty, not able to calibrate"
            )
        if len(ptr_mean) == 0:
            raise CalibrationError(
                "The predicted tr list is empty, not able to calibrate"
            )

        # calculate calibration curves
        for i in range(0, len(ptr_mean)):
            if i >= len(ptr_mean) - 1:
                continue
            delta_ptr = ptr_mean[i + 1] - ptr_mean[i]
            delta_mtr = mtr_mean[i + 1] - mtr_mean[i]

            slope = delta_mtr / delta_ptr
            intercept = (-1 * (ptr_mean[i] * slope)) + mtr_mean[i]

            # optimized predictions using a dict to find calibration curve very
            # fast
            for v in np.arange(
                round(ptr_mean[i], self.bin_dist),
                round(ptr_mean[i + 1], self.bin_dist),
                1 / ((self.bin_dist) * self.dict_cal_divider),
            ):
                if v < calibrate_min:
                    calibrate_min = v
                if v > calibrate_max:
                    calibrate_max = v
                calibrate_dict[str(round(v, self.bin_dist))] = [slope, intercept]

        return calibrate_min, calibrate_max, calibrate_dict

    def calibrate_preds(self,
                        psm_list=None,
                        infile="",
                        measured_tr=[],
                        correction_factor=1.0,
                        location_retraining_models="",
                        psm_utils_obj=None,
                        sample_for_calibration_curve=None,
                        seq_df=None,
                        use_median=True,
                        return_plotly_report=False):
        """
        Find best model and calibrate.

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : list
            identifiers of the peptides; should correspond to seqs and mods
        measured_tr : list
            measured tr of the peptides; should correspond to seqs, identifiers,
            and mods
        correction_factor : float
            correction factor that needs to be applied to the supplied measured
            trs
        seq_df : object :: pd.DataFrame
            a pd.DataFrame that contains the sequences, modifications and
            observed retention times to fit a calibration curve
        use_median : boolean
            flag to indicate we need to use the median valuein a window to
            perform calibration

        Returns
        -------

        """
        if type(seq_df) == pd.core.frame.DataFrame:
            list_of_psms = []
            if self.predict_ccs:
                for seq,mod,ident,tr,z in zip(seq_df["seq"],seq_df["modifications"],seq_df.index,seq_df["tr"],seq_df["charge"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod,charge=z),spectrum_id=ident,retention_time=tr))
            else:
                for seq,mod,ident,tr in zip(seq_df["seq"],seq_df["modifications"],seq_df.index,seq_df["tr"]):
                    list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident,retention_time=tr))
            psm_list = PSMList(psm_list=list_of_psms)
        elif psm_utils_obj:
            psm_list = psm_utils_obj    

        if isinstance(self.model, str):
            self.model = [self.model]
        
        if len(infile) > 0:
            psm_list = read_file(infile)
            if "msms" in infile and ".txt" in infile:
                mapper = pd.read_csv(os.path.join(os.path.dirname(os.path.realpath(__file__)), "unimod/map_mq_file.csv"),index_col=0)["value"].to_dict()
                psm_list.rename_modifications(mapper)

        measured_tr = [psm.retention_time for psm in psm_list]

        if self.verbose:
            logger.debug("Start to calibrate predictions ...")
        if self.verbose:

            logger.debug(
                "Ready to find the best model out of: %s" %
                (self.model))
        
        best_perf = float("inf")
        best_calibrate_min = 0.0
        best_calibrate_max = 0.0
        best_calibrate_dict = {}
        mod_calibrate_dict = {}
        mod_calibrate_min_dict = {}
        mod_calibrate_max_dict = {}
        pred_dict = {}
        mod_dict = {}
        temp_obs = []
        temp_pred = []

        if self.deeplc_retrain:
            # The following code is not required in most cases, but here it is used to clear variables that might cause problems
            _ = tf.Variable([1])

            context._context = None
            context._create_context()

            tf.config.threading.set_inter_op_parallelism_threads(1)

            if len(location_retraining_models) > 0:
                t_dir_models = TemporaryDirectory().name
                os.mkdir(t_dir_models)
            else:
                t_dir_models = location_retraining_models
                try:
                    os.mkdir(t_dir_models)
                except:
                    pass

            # Here we will apply transfer learning we specify previously trained models in the 'mods_transfer_learning'
            models = deeplcretrainer.retrain(
                {"deeplc_transferlearn":psm_list},
                outpath=t_dir_models,
                mods_transfer_learning=self.model,
                freeze_layers=True,
                n_epochs=self.n_epochs,
                freeze_after_concat=1,
            )

            self.model = models

        if isinstance(sample_for_calibration_curve, int):
            psm_list = random.sample(list(psm_list), sample_for_calibration_curve)
            measured_tr = [psm.retention_time for psm in psm_list]

        for m in self.model:
            if self.verbose:
                logger.debug("Trying out the following model: %s" % (m))
            if self.pygam_calibration:
                calibrate_output = self.calibrate_preds_func_pygam(
                    psm_list,
                    measured_tr=measured_tr,
                    correction_factor=correction_factor,
                    seq_df=seq_df,
                    use_median=use_median,
                    mod_name=m,
                )
            else:
                calibrate_output = self.calibrate_preds_func(
                    psm_list,
                    correction_factor=correction_factor,
                    seq_df=seq_df,
                    use_median=use_median,
                    mod_name=m,
                )

            (
                self.calibrate_min,
                self.calibrate_max,
                self.calibrate_dict,
            ) = calibrate_output

            if type(self.calibrate_dict) == dict:
                if len(self.calibrate_dict.keys()) == 0:
                    continue
            
            m_name = m.split("/")[-1]

            preds = self.make_preds(psm_list,
                                    calibrate=True,
                                    seq_df=seq_df,
                                    mod_name=m)

            if self.deepcallc_mod:
                m_group_name = "deepcallc"
            else:
                m_group_name = "_".join(m_name.split("_")[:-1])

            try:
                pred_dict[m_group_name][m] = preds
                mod_dict[m_group_name][m] = m
                mod_calibrate_dict[m_group_name][m] = self.calibrate_dict
                mod_calibrate_min_dict[m_group_name][m] = self.calibrate_min
                mod_calibrate_max_dict[m_group_name][m] = self.calibrate_max
            except KeyError:
                pred_dict[m_group_name] = {}
                mod_dict[m_group_name] = {}
                mod_calibrate_dict[m_group_name] = {}
                mod_calibrate_min_dict[m_group_name] = {}
                mod_calibrate_max_dict[m_group_name] = {}

                pred_dict[m_group_name][m] = preds
                mod_dict[m_group_name][m] = m
                mod_calibrate_dict[m_group_name][m] = self.calibrate_dict
                mod_calibrate_min_dict[m_group_name][m] = self.calibrate_min
                mod_calibrate_max_dict[m_group_name][m] = self.calibrate_max

        for m_name in pred_dict.keys():
            preds = [sum(a) / len(a) for a in zip(*list(pred_dict[m_name].values()))]
            if len(measured_tr) == 0:
                perf = sum(abs(seq_df["tr"] - preds))
            else:
                perf = sum(abs(np.array(measured_tr) - np.array(preds)))

            if self.verbose:
                logger.debug(
                    "For %s model got a performance of: %s"
                    % (m_name, perf / len(preds))
                )

            if perf < best_perf:
                if self.deepcallc_mod:
                    m_group_name = "deepcallc"
                else:
                    m_group_name = m_name
                    # TODO is deepcopy really required?

                best_calibrate_dict = copy.deepcopy(mod_calibrate_dict[m_group_name])
                best_calibrate_min = copy.deepcopy(mod_calibrate_min_dict[m_group_name])
                best_calibrate_max = copy.deepcopy(mod_calibrate_max_dict[m_group_name])

                best_model = copy.deepcopy(mod_dict[m_group_name])
                best_perf = perf

                temp_obs = np.array(measured_tr)
                temp_pred = np.array(preds)

        self.calibrate_dict = best_calibrate_dict
        self.calibrate_min = best_calibrate_min
        self.calibrate_max = best_calibrate_max
        self.model = best_model

        if self.deepcallc_mod:
            self.deepcallc_model = train_en(pd.DataFrame(pred_dict["deepcallc"]),seq_df["tr"])

        self.n_jobs = 1

        logger.debug("Model with the best performance got selected: %s" % (best_model))

        if return_plotly_report:
            import deeplc.plot
            plotly_return_dict = {}
            plotly_df = pd.DataFrame(
                            list(zip(temp_obs,temp_pred)),
                            columns=["Observed retention time","Predicted retention time"]
                        )
            plotly_return_dict["scatter"] = deeplc.plot.scatter(plotly_df)
            plotly_return_dict["baseline_dist"] = deeplc.plot.distribution_baseline(plotly_df)
            return plotly_return_dict

        return {}

    def split_seq(self, a, n):
        """
        Split a list (a) into multiple chunks (n)

        Parameters
        ----------
        a : list
            list to split
        n : list
            number of chunks

        Returns
        -------
        list
            chunked list
        """

        # since chunking is not alway possible do the modulo of residues
        k, m = divmod(len(a), n)
        result = (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))
        return result



"""
This code is used to extract features for peptides used in DeepLC (or seperately
if required).

For the library versions see the .yml file
"""

__author__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__credits__ = ["Robbin Bouwmeester", "Ralf Gabriels", "Arthur Declercq", "Prof. Lennart Martens", "Sven Degroeve"]
__license__ = "Apache License, Version 2.0"
__maintainer__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__email__ = ["Robbin.Bouwmeester@ugent.be", "Ralf.Gabriels@ugent.be"]

# Native imports
from operator import index
import os
import math
import time
from configparser import ConfigParser
import ast
from re import sub
import logging
from copy import deepcopy

# Numpy
import numpy as np

# Pandas
import pandas as pd

from psm_utils.io.peptide_record import peprec_to_proforma
from psm_utils.psm import PSM
from psm_utils.psm_list import PSMList
from pyteomics import mass

from functools import lru_cache

from psm_utils.io.peptide_record import peprec_to_proforma
from psm_utils.psm import PSM
from psm_utils.psm_list import PSMList
from psm_utils.io import read_file
from psm_utils.io import write_file

logger = logging.getLogger(__name__)


class FeatExtractor():
    """
    Place holder, fill later

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self,
                 main_path=os.path.dirname(os.path.realpath(__file__)),
                 lib_path_mod=os.path.join(os.path.dirname(os.path.realpath(__file__)), "unimod/"),
                 lib_aa_composition=os.path.join(os.path.dirname(os.path.realpath(__file__)), "aa_comp_rel.csv"),
                 split_size=7,
                 verbose=True,
                 include_specific_posses=[0, 1, 2, 3, 4, 5, 6, -1, -2, -3, -4, -5, -6, -7],
                 add_sum_feat=False,
                 ptm_add_feat=False,
                 ptm_subtract_feat=False,
                 add_rolling_feat=False,
                 include_unnormalized=True,
                 add_comp_feat=False,
                 cnn_feats=True,
                 ignore_mods=False,
                 config_file=None):
        # if a config file is defined overwrite standard parameters
        if config_file:
            cparser = ConfigParser()
            cparser.read(config_file)
            lib_path_mod = cparser.get(
                "featExtractor", "lib_path_mod").strip('"')
            split_size = cparser.getint("featExtractor", "split_size")
            verbose = cparser.getboolean("featExtractor", "verbose")
            add_sum_feat = cparser.getboolean("featExtractor", "add_sum_feat")
            ptm_add_feat = cparser.getboolean("featExtractor", "ptm_add_feat")
            ptm_subtract_feat = cparser.getboolean(
                "featExtractor", "ptm_subtract_feat")
            add_rolling_feat = cparser.getboolean(
                "featExtractor", "add_rolling_feat")
            include_unnormalized = cparser.getboolean(
                "featExtractor", "include_unnormalized")
            include_specific_posses = ast.literal_eval(
                cparser.get("featExtractor", "include_specific_posses"))

        self.main_path = main_path
        self.lib_add, self.lib_subtract = self.get_libs_mods(lib_path_mod)
        self.lib_add = dict([(k.lower(), v) for k, v in self.lib_add.items()])
        self.lib_subtract = dict([(k.lower(), v)
                                  for k, v in self.lib_subtract.items()])

        # Get the atomic composition of AAs
        self.lib_aa_composition = self.get_aa_composition(lib_aa_composition)

        self.split_size = split_size
        self.verbose = verbose

        self.add_sum_feat = add_sum_feat
        self.ptm_add_feat = ptm_add_feat
        self.ptm_subtract_feat = ptm_subtract_feat
        self.add_rolling_feat = add_rolling_feat
        self.cnn_feats = cnn_feats
        self.include_unnormalized = include_unnormalized
        self.include_specific_posses = include_specific_posses
        self.add_comp_feat = add_comp_feat
        self.ignore_mods = ignore_mods

    def __str__(self):
        return("""
  _____                  _      _____       __           _               _                  _
 |  __ \                | |    / ____|     / _|         | |             | |                | |
 | |  | | ___  ___ _ __ | |   | |   ______| |_ ___  __ _| |_    _____  _| |_ _ __ __ _  ___| |_ ___  _ __
 | |  | |/ _ \/ _ \ '_ \| |   | |  |______|  _/ _ \/ _` | __|  / _ \ \/ / __| '__/ _` |/ __| __/ _ \| '__|
 | |__| |  __/  __/ |_) | |___| |____     | ||  __/ (_| | |_  |  __/>  <| |_| | | (_| | (__| || (_) | |
 |_____/ \___|\___| .__/|______\_____|    |_| \___|\__,_|\__|  \___/_/\_\\__|_|  \__,_|\___|\__\___/|_|
                  | |
                  |_|
        """)

    def get_aa_composition(self, file_loc):
        """
        Read amino acid atomic composition and return a dictionary

        Parameters
        ----------
        file_loc : str
            location of the (csv) file that contains the atomic compositions of AAs. The first column must contain
            the one-letter AA code. The remaining columns contain counts for each atom (each atom in seperate
            column). An example is:

                aa,C,H,N,O,S
                A,1,2,0,0,0
                R,4,9,3,0,0
                N,2,3,1,1,0

        Returns
        -------
        dict
            dictionary that goes from one-letter AA to atomic composition
        """
        return(pd.read_csv(file_loc, index_col=0).T.to_dict())

    def split_seq(self,
                  a,
                  n):
        """
        Split a list (a) into multiple chunks (n)

        Parameters
        ----------
        a : list
            list to split
        n : list
            number of chunks

        Returns
        -------
        list
            chunked list
        """
        # since chunking is not alway possible do the modulo of residues
        k, m = divmod(len(a), n)
        return(a[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))

    def get_libs_mods(self, directory):
        """
        Make a dictionary with unimod to chemical formula

        Parameters
        ----------
        directory : str
            directory of the unimod to chemical formula mapping

        Returns
        -------
        dict
            chemical formula of a PTM when it is added
        dict
            chemical formula of a PTM when it is subtracted
        """
        # TODO replace dir with actual file...
        mod_df = pd.read_csv(
            os.path.join(
                directory,
                "unimod_to_formula.csv"),
            index_col=0)
        mod_dict = mod_df.to_dict()
        return mod_dict["formula_pos"], mod_dict["formula_neg"]

    def calc_feats_mods(self,
                        formula):
        """
        Chemical formula to atom addition/subtraction

        Parameters
        ----------
        formula : str
            chemical formula

        Returns
        -------
        list
            atom naming
        list
            number of atom added/subtracted
        """
        if not formula:
            return [], []
        if len(str(formula)) == 0:
            return [], []
        if not isinstance(formula, str):
            if math.isnan(formula):
                return [], []

        new_atoms = []
        new_num_atoms = []
        for atom in formula.split(" "):
            if "(" not in atom:
                atom_symbol = atom
                num_atom = 1
            else:
                atom_symbol = atom.split("(")[0]
                num_atom = atom.split("(")[1].rstrip(")")
            new_atoms.append(atom_symbol)
            new_num_atoms.append(int(num_atom))
        return new_atoms, new_num_atoms

    def get_feats_mods(self,
                       seqs,
                       mods,
                       identifiers,
                       split_size=False,
                       atoms_order=set(["H", "C", "N", "O", "P", "S"]),
                       add_str="_sum",
                       subtract_mods=False):
        """
        Chemical formula to atom addition/subtraction

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : str
            identifiers of the peptides; should correspond to seqs and mods
        split_size : int
            overwrite the set split size if needed
        atoms_order : set
            atoms to include and the order
        add_str : str
            add this substring to feature naming
        subtract_mods : boolean
            calculate the atom that are substracted in the PTM

        Returns
        -------
        object :: pd.DataFrame
            feature matrix for peptide PTMs
        """
        if not split_size:
            split_size = self.split_size
        if self.verbose:
            t0 = time.time()
        mod_dict = {}
        look_up_mod_subtract = {}
        look_up_mod_add = {}

        len_init = len([ao + str(spl_s)
                        for spl_s in range(split_size) for ao in atoms_order])
        for index_name, mod, seq in zip(identifiers, mods, seqs):
            mod_dict[index_name] = dict(zip(
                [ao + str(spl_s) + add_str for spl_s in range(split_size) for ao in atoms_order], [0] * len_init))

            if not mod:
                continue
            if len(str(mod)) == 0:
                continue
            if not isinstance(mod, str):
                if math.isnan(mod):
                    continue
            """
            split_mod = mod.rstrip().split("|")
            for i in range(1, len(split_mod), 2):
                if subtract_mods:
                    try:
                        fill_mods, num = look_up_mod_subtract[self.lib_subtract[split_mod[i]]]
                    except KeyError:
                        look_up_mod_subtract[self.lib_subtract[split_mod[i]]] = self.calc_feats_mods(
                            self.lib_subtract[split_mod[i]])
                        fill_mods, num = look_up_mod_subtract[self.lib_subtract[split_mod[i]]]
                else:
                    try:
                        fill_mods, num = look_up_mod_add[self.lib_subtract[split_mod[i]]]
                    except KeyError:
                        look_up_mod_add[self.lib_subtract[split_mod[i]]] = self.calc_feats_mods(
                            self.lib_add[split_mod[i]])
                        fill_mods, num = look_up_mod_add[self.lib_subtract[split_mod[i]]]

                loc = int(split_mod[i - 1])

                if loc > len(seq):
                    loc = len(seq)

                relative_loc = int(
                    math.ceil(
                        (loc / len(seq)) * split_size)) - 1
                for fm, n in zip(fill_mods, num):
                    if fm not in atoms_order:
                        continue
                    mod_dict[index_name]["%s%s%s" %
                                         (fm, relative_loc, add_str)] += n
            """

        if self.verbose:
            logger.debug(
                "Time to calculate mod features: %s seconds" %
                (time.time() - t0))
        df_ret = pd.DataFrame(mod_dict, dtype=int).T
        del mod_dict
        return df_ret

    def encode_atoms(self,
                     psm_list,
                     indexes,
                     charges=[],
                     predict_ccs=False,
                     padding_length=60,
                     positions=set([0, 1, 2, 3, -1, -2, -3, -4]),
                     positions_pos=set([0, 1, 2, 3]),
                     positions_neg=set([-1, -2, -3, -4]),
                     sum_mods=2,
                     dict_aa={
                         "K": 0,
                         "R": 1,
                         "P": 2,
                         "T": 3,
                         "N": 4,
                         "A": 5,
                         "Q": 6,
                         "V": 7,
                         "S": 8,
                         "G": 9,
                         "I": 10,
                         "L": 11,
                         "C": 12,
                         "M": 13,
                         "H": 14,
                         "F": 15,
                         "Y": 16,
                         "W": 17,
                         "E": 18,
                         "D": 19
                     },
                     dict_index_pos={'C': 0,
                                     'H': 1,
                                     'N': 2,
                                     'O': 3,
                                     'S': 4,
                                     'P': 5},
                     dict_index_all={'C': 0,
                                     'H': 1,
                                     'N': 2,
                                     'O': 3,
                                     'S': 4,
                                     'P': 5},
                     dict_index={'C': 0,
                                 'H': 1,
                                 'N': 2,
                                 'O': 3,
                                 'S': 4,
                                 'P': 5}):
        """
        Extract all features we can extract... Probably the function your want to call by default

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods_all : list
            naming of the mods; should correspond to seqs and identifiers
        indexes : list
            identifiers of the peptides; should correspond to seqs and mods
        padding_length : int
            indicates padding length with 'X'-characters. Shorter sequences are padded. Longer sequences
            are sliced shorter (C-terminal > than padding length will be missing)
        positions : list
            list of positions to include seperately, for the C-terminus
            provide negative indices
        sum_mods : int
            value that is used to feed the second head of cerberus with summed information, for example,
            a value of 2 would sum the composition of each 2 AAs (i.e. 1+2, 3+4, 5+6 ...)
        dict_index_pos : dict
            index position of atom for positional features
        dict_index_all : dict
            index position of atom for overall compositional features
        dict_index : dict
            index position of atom for compositional features for the whole peptide (each position)
        charges : list
            optional list with charges, keep emtpy if these will not effect the predicted value

        Returns
        -------
        object :: pd.DataFrame
            feature matrix (np.matrix) of all positions (up till padding length)
        object :: pd.DataFrame
            feature matrix (np.matrix) of summed positions (up till paddint length / sum_mods)
        object :: pd.DataFrame
            feature matrix (np.matrix) of specific positions (from positions argument)
        object :: pd.DataFrame
            feature matrix (np.matrix) of summed composition
        """
        # TODO param flag for CCS prediction
        def rolling_sum(a, n=2):
            ret = np.cumsum(a, axis=1, dtype=np.float32)
            ret[:, n:] = ret[:, n:] - ret[:, :-n]
            return ret[:, n - 1:]
        
        ret_list = {}
        ret_list["matrix"] = {}
        ret_list["matrix_sum"] = {}
        ret_list["matrix_all"] = {}
        ret_list["pos_matrix"] = {}
        ret_list["matrix_hc"] = {}

        # Iterate over all instances
        for psm,row_index in zip(psm_list,indexes):
            peptidoform = psm.peptidoform
            seq = peptidoform.sequence
            seq_len = len(seq)
            charge = psm.get_precursor_charge()
            
            # For now anything longer than padding length is cut away
            # (C-terminal cutting)
            if seq_len > padding_length:
                seq = seq[0:padding_length]
                seq_len = len(seq)
                logger.warning(
                            "Truncating peptide (too long): %s" % (seq))

            peptide_composition = [mass.std_aa_comp[aa] for aa in seq]

            # Initialize all feature matrixes
            matrix = np.zeros(
                (padding_length, len(dict_index.keys())), dtype=np.float16)
            matrix_hc = np.zeros(
                (padding_length, len(dict_aa.keys())), dtype=np.float16)
            matrix_pos = np.zeros(
                (len(positions), len(
                    dict_index.keys())), dtype=np.float16)
            
            for i, position_composition in enumerate(peptide_composition):
                for k, v in position_composition.items():
                    try:
                        matrix[i, dict_index[k]] = v
                    except IndexError:
                        logger.warning(f"Could not add the following value: pos {i} for atom {k} with value {v}")
                    except KeyError:
                        logger.warning(f"Could not add the following value: pos {i} for atom {k} with value {v}")

            for p in positions_pos:
                aa = seq[p]
                for atom, val in mass.std_aa_comp[aa].items():
                    try:
                        matrix_pos[p, dict_index_pos[atom]] = val
                    except KeyError:
                        logger.warning(f"Could not add the following atom: {atom}")
                    except IndexError:
                        logger.warning(f"Could not add the following atom: {p} {atom} {val}")

            for pn in positions_neg:
                aa = seq[seq_len + pn]
                for atom, val in mass.std_aa_comp[aa].items():
                    try:
                        matrix_pos[pn, dict_index_pos[atom]] = val
                    except KeyError:
                        logger.warning(f"Could not add the following atom: {atom}")
                    except IndexError:
                        logger.warning(f"Could not add the following atom: {pn} {atom} {val}")

            for i, peptide_position in enumerate(peptidoform.parsed_sequence):
                try:
                    matrix_hc[i, dict_aa[peptide_position[0]]] = 1.
                except KeyError:
                    logger.warning(
                            f"Skipping the following (not in library): {i} {peptide_position}")
                except IndexError:
                    # Likely to be a sequence > 60 AA
                    logger.warning(f"Could not add the following atom: {i} {peptide_position}")

                if peptide_position[1] is not None:
                    try:
                        modification_composition = peptide_position[1][0].composition
                    except KeyError:
                        logger.warning(
                            f"Skipping the following (not in library): {peptide_position[1]}")
                        continue
                    except:
                        logger.warning(
                            f"Skipping the following (not in library): {peptide_position[1]}")
                        continue

                    for atom_position_composition,atom_change in modification_composition.items():
                        try:
                            matrix[i, dict_index[atom_position_composition]] += atom_change
                            if i in positions:
                                matrix_pos[i, dict_index_pos[atom_position_composition]] += atom_change
                            elif i - seq_len in positions:
                                matrix_pos[i - seq_len, dict_index_pos[atom_position_composition]] += atom_change
                        except KeyError:
                            try:
                                logger.warning(f"Could not add the following atom: {atom_position_composition}, attempting to replace the [] part")
                                atom_position_composition = sub("\[.*?\]", "", atom_position_composition)
                                matrix[i, dict_index[atom_position_composition]] += atom_change
                                if i in positions:
                                    matrix_pos[i, dict_index_pos[atom_position_composition]] += atom_change
                                elif i - seq_len in positions:
                                    matrix_pos[i - seq_len, dict_index_pos[atom_position_composition]] += atom_change
                            except KeyError:
                                logger.warning(f"Could not add the following atom: {atom_position_composition}, second attempt, now ignored")
                                continue
                            except:
                                logger.warning(f"Could not add the following atom: {atom_position_composition}, second attempt, now ignored")
                                continue
                        except IndexError:
                            logger.warning(f"Could not add the following atom: {i} {atom_position_composition} {atom_change}")
                        except:
                            logger.warning(f"Could not add the following atom: {atom_position_composition}, second attempt, now ignored")
                            continue


            matrix_all = np.sum(matrix, axis=0)
            matrix_all = np.append(matrix_all, seq_len)

            if predict_ccs:
                matrix_all = np.append(matrix_all,(seq.count("H"))/float(seq_len))
                matrix_all = np.append(matrix_all,(seq.count("F")+seq.count("W")+seq.count("Y"))/float(seq_len))
                matrix_all = np.append(matrix_all,(seq.count("D")+seq.count("E"))/float(seq_len))
                matrix_all = np.append(matrix_all,(seq.count("K")+seq.count("R"))/float(seq_len))
                matrix_all = np.append(matrix_all,charge)

            matrix_sum = rolling_sum(matrix.T, n=2)[:, ::2].T

            ret_list["matrix"][row_index] = matrix 
            ret_list["matrix_sum"][row_index] = matrix_sum
            ret_list["matrix_all"][row_index] = matrix_all
            ret_list["pos_matrix"][row_index] = matrix_pos.flatten()
            ret_list["matrix_hc"][row_index] = matrix_hc

        return ret_list
    
    
    def full_feat_extract(self,
                          psm_list=[],
                          seqs=[],
                          mods=[],
                          predict_ccs=False,
                          identifiers=[],
                          charges=[]):
        """
        Extract all features we can extract... Probably the function your want to call by default

        Parameters
        ----------
        seqs : list
            peptide sequence list; should correspond to mods and identifiers
        mods : list
            naming of the mods; should correspond to seqs and identifiers
        identifiers : str
            identifiers of the peptides; should correspond to seqs and mods
        charges : list
            optional list with charges, keep emtpy if these will not effect the predicted value

        Returns
        -------
        pd.DataFrame
            feature matrix
        """
        # TODO Reintroduce for CCS, check CCS flag
        if len(seqs) > 0:
            list_of_psms = []
            for seq,mod,id in zip(seqs,mods,identifiers):
                list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=id))
            psm_list = PSMList(psm_list=list_of_psms)

        if self.verbose:
            t0 = time.time()

        # TODO pass CCS flag
        if self.add_sum_feat:
            if self.verbose:
                logger.debug(
                    "Extracting compositional sum features for modifications")
            X_feats_sum = self.get_feats_mods(
                psm_list, split_size=1, add_str="_sum")
        if self.ptm_add_feat:
            if self.verbose:
                logger.debug(
                    "Extracting compositional add features for modifications")
            X_feats_add = self.get_feats_mods(
                psm_list, split_size=self.split_size, add_str="_add")
        if self.ptm_subtract_feat:
            if self.verbose:
                logger.debug(
                    "Extracting compositional subtract features for modifications")
            X_feats_neg = self.get_feats_mods(
                psm_list,
                split_size=self.split_size,
                add_str="_subtract",
                subtract_mods=True)
        if self.cnn_feats:
            if self.verbose:
                logger.debug("Extracting CNN features")
            X_cnn = self.encode_atoms(
                psm_list, list(range(len(psm_list))), charges=charges, predict_ccs=predict_ccs)

        if self.cnn_feats:
            X = X_cnn
        if self.add_sum_feat:
            try:
                X = pd.concat([X, X_feats_sum], axis=1)
            except BaseException:
                X = X_feats_sum
        if self.ptm_add_feat:
            try:
                X = pd.concat([X, X_feats_add], axis=1)
            except BaseException:
                X = X_feats_add
        if self.ptm_subtract_feat:
            try:
                X = pd.concat([X, X_feats_neg], axis=1)
            except BaseException:
                X = X_feats_neg

        if self.verbose:
            logger.debug(
                "Time to calculate all features: %s seconds" %
                (time.time() - t0))
        return X


def main(verbose=True):
    f_extractor = FeatExtractor(config_file="config.ini")
    df = pd.read_csv("parse_pride/seqs_exp.csv")
    df.index = ["Pep_" + str(dfi) for dfi in df.index]
    print(
        f_extractor.full_feat_extract(
            df["seq"],
            df["modifications"],
            df.index))


if __name__ == "__main__":
    main()



"""Graphical user interface."""
import sys
from multiprocessing import freeze_support
from pathlib import Path
import importlib.resources
try:
    from gooey import Gooey, local_resource_path
except ImportError:
    raise ImportError(
        "Missing dependency `gooey` required to start graphical user interface. "
        "Install `gooey` or use the command line interface."
    )

import deeplc.package_data.gui_images as img_module
from deeplc.__main__ import main


# Get path to package_data/images
# Workaround with parent of specific file required for Python 3.9+ support
with importlib.resources.path(img_module, 'config_icon.png') as resource:
    _IMG_DIR = Path(resource).parent


@Gooey(
    program_name="DeepLC",
    image_dir=local_resource_path(_IMG_DIR),
    tabbed_groups=True,
    default_size=(720, 480),
    monospace_display=True,
    target=None if getattr(sys, 'frozen', False) else "deeplc-gui"
)
def start_gui():
    """Run main with GUI enabled."""
    freeze_support()  # Required for multiprocessing with PyInstaller
    main(gui=True)

if __name__ == "__main__":
    start_gui()



"""
Robbin Bouwmeester

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
This code is used to train retention time predictors and store
predictions from a CV procedure for further analysis.

This project was made possible by MASSTRPLAN. MASSTRPLAN received funding 
from the Marie Sklodowska-Curie EU Framework for Research and Innovation 
Horizon 2020, under Grant Agreement No. 675132.
"""

from sklearn.model_selection import RandomizedSearchCV
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error
from sklearn.feature_selection import SelectFromModel
from sklearn.model_selection import cross_val_predict
from sklearn.model_selection import KFold
from sklearn.base import clone
from sklearn.model_selection import GridSearchCV
from scipy.stats import randint
from scipy.stats import uniform
from numpy import arange
from scipy.stats import pearsonr

from operator import itemgetter
from numpy import median
from collections import Counter
    
def train_en(X,y,n_jobs=16,cv=None):
    """
    Function that trains Layer 3 of CALLC (elastic net)
    
    Parameters
    ----------
    X : pd.DataFrame
        dataframe with molecular descriptors
    y : pd.Series
        vector with observed retention times
    n_jobs : int
        number of jobs to spawn
    cv : sklearn.model_selection.KFold
        cv object
    
    Returns
    -------
    sklearn.linear_model.ElasticNet
        elastic net model trained in Layer 3
    list
        list with predictions
    list
        list with features used to train Layer 3
    """
    preds = []

    model = ElasticNet()
    crossv_mod = clone(model)
    ret_mod = clone(model)

    set_reg = [0.01,1.0,10.0,100.0,1000.0,10000.0,10000.0,100000.0,1000000.0,1000000000,1000000]
    set_reg.extend([x/2 for x in set_reg])
    set_reg.extend([x/3 for x in set_reg])
    
    params = {
       'alpha': set_reg,
       'l1_ratio' : [0.01,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0],
       'copy_X':[True],
       'normalize' : [False],
       'positive' : [True],
       'fit_intercept'  : [True,False]
    }

    grid = GridSearchCV(model, params,cv=cv,scoring='neg_mean_absolute_error',verbose=0,n_jobs=n_jobs,refit=True)
    grid.fit(X,y)
    
    cv_pred = cv
    crossv_mod.set_params(**grid.best_params_)
    preds = cross_val_predict(crossv_mod, X=X, y=y, cv=cv_pred, n_jobs=n_jobs, verbose=0)

    ret_mod.set_params(**grid.best_params_)
    ret_mod.fit(X,y)

    coef_indexes = [i for i,coef in enumerate(ret_mod.coef_) if coef > 0.0]
   
    return ret_mod


import sys


if sys.version_info >= (3,8):
    from importlib.metadata import version
    __version__ = version('deeplc')
else:
    import pkg_resources
    __version__ = pkg_resources.require("deeplc")[0].version


from deeplc.deeplc import DeepLC
from deeplc.feat_extractor import FeatExtractor




"""Main command line interface to DeepLC."""

__author__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__credits__ = ["Robbin Bouwmeester", "Ralf Gabriels", "Prof. Lennart Martens", "Sven Degroeve"]
__license__ = "Apache License, Version 2.0"
__maintainer__ = ["Robbin Bouwmeester", "Ralf Gabriels"]
__email__ = ["Robbin.Bouwmeester@ugent.be", "Ralf.Gabriels@ugent.be"]

import logging
import os
import sys
import warnings

import pandas as pd
from matplotlib import pyplot as plt

from deeplc import __version__, DeepLC, FeatExtractor
from deeplc._argument_parser import parse_arguments
from deeplc._exceptions import DeepLCError

from psm_utils.io.peptide_record import peprec_to_proforma
from psm_utils.psm import PSM
from psm_utils.psm_list import PSMList
from psm_utils.io import read_file
from psm_utils.io import write_file

logger = logging.getLogger(__name__)


def setup_logging(passed_level):
    log_mapping = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }

    if passed_level.lower() not in log_mapping:
        print(
            "Invalid log level. Should be one of the following: ",
            ', '.join(log_mapping.keys())
        )
        exit(1)

    logging.basicConfig(
        stream=sys.stdout,
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        level=log_mapping[passed_level.lower()]
    )

def main(gui=False):
    """Main function for the CLI."""
    argu = parse_arguments(gui=gui)

    setup_logging(argu.log_level)

    # Reset logging levels if DEBUG (see deeplc.py)
    if argu.log_level.lower() == "debug":
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
        logging.getLogger('tensorflow').setLevel(logging.DEBUG)
        warnings.filterwarnings('default', category=DeprecationWarning)
        warnings.filterwarnings('default', category=FutureWarning)
        warnings.filterwarnings('default', category=UserWarning)
    else:
        os.environ['KMP_WARNINGS'] = '0'

    try:
        run(**vars(argu))
    except DeepLCError as e:
        logger.exception(e)
        sys.exit(1)


def run(
    file_pred,
    file_cal=None,
    file_pred_out=None,
    file_model=None,
    pygam_calibration=True,
    split_cal=50,
    dict_divider=50,
    use_library=None,
    write_library=False,
    batch_num=50000,
    n_threads=None,
    transfer_learning=False,
    log_level="info",
    verbose=True,
):
    """Run DeepLC."""
    logger.info("Using DeepLC version %s", __version__)
    logger.debug("Using %i CPU threads", n_threads)

    df_pred = False
    df_cal = False
    first_line_pred = ""
    first_line_cal = ""

    if not file_cal and file_model != None:
        fm_dict = {}
        sel_group = ""
        for fm in file_model:
            if len(sel_group) == 0:
                sel_group = "_".join(fm.split("_")[:-1])
                fm_dict[sel_group]= fm
                continue
            m_group = "_".join(fm.split("_")[:-1])
            if m_group == sel_group:
                fm_dict[m_group] = fm
        file_model = fm_dict
    
    with open(file_pred) as f:
        first_line_pred = f.readline().strip()
    if file_cal:
        with open(file_cal) as f:
            first_line_cal = f.readline().strip()

    if "modifications" in first_line_pred.split(",") and "seq" in first_line_pred.split(","):
        # Read input files
        df_pred = pd.read_csv(file_pred)
        if len(df_pred.columns) < 2:
            df_pred = pd.read_csv(file_pred,sep=" ")
        df_pred = df_pred.fillna("")
        file_pred = ""

        list_of_psms = []
        for seq,mod,ident in zip(df_pred["seq"],df_pred["modifications"],df_pred.index):
            list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident))
        psm_list_pred = PSMList(psm_list=list_of_psms)
        df_pred = None
    else:
        psm_list_pred = read_file(file_pred)
        if "msms" in file_pred and ".txt" in file_pred:
            mapper = pd.read_csv(os.path.join(os.path.dirname(os.path.realpath(__file__)), "unimod/map_mq_file.csv"),index_col=0)["value"].to_dict()
            psm_list_pred.rename_modifications(mapper)

    # Allow for calibration file to be empty (undefined), fill in if/elif if present
    psm_list_cal = []
    if "modifications" in first_line_cal.split(",") and "seq" in first_line_cal.split(",") and file_cal:
        df_cal = pd.read_csv(file_cal)
        if len(df_cal.columns) < 2:
            df_cal = pd.read_csv(df_cal,sep=" ")
        df_cal = df_cal.fillna("")
        file_cal = ""

        list_of_psms = []
        for seq,mod,ident,tr in zip(df_cal["seq"],df_cal["modifications"],df_cal.index,df_cal["tr"]):
            list_of_psms.append(PSM(peptidoform=peprec_to_proforma(seq,mod),spectrum_id=ident,retention_time=tr))
        psm_list_cal = PSMList(psm_list=list_of_psms)
        df_cal = None
    elif file_cal:
        psm_list_cal = read_file(file_cal)
        if "msms" in file_cal and ".txt" in file_cal:
            mapper = pd.read_csv(os.path.join(os.path.dirname(os.path.realpath(__file__)), "unimod/map_mq_file.csv"),index_col=0)["value"].to_dict()
            psm_list_cal.rename_modifications(mapper)
    # Make a feature extraction object; you can skip this if you do not want to
    # use the default settings for DeepLC. Here we want to use a model that does
    # not use RDKit features so we skip the chemical descriptor making
    # procedure.
    f_extractor = FeatExtractor(
        cnn_feats=True,
        verbose=verbose
    )
    
    # Make the DeepLC object that will handle making predictions and calibration
    dlc = DeepLC(
        path_model=file_model,
        f_extractor=f_extractor,
        cnn_model=True,
        split_cal=split_cal,
        dict_cal_divider=dict_divider,
        write_library=write_library,
        use_library=use_library,
        batch_num=batch_num,
        n_jobs=n_threads,
        verbose=verbose,
        deeplc_retrain=transfer_learning
    )
    
    # Calibrate the original model based on the new retention times
    if len(psm_list_cal) > 0:
        logger.info("Selecting best model and calibrating predictions...")
        logger.info("Initiating transfer learning?")
        dlc.calibrate_preds(psm_list=psm_list_cal)

    # Make predictions; calibrated or uncalibrated
    logger.info("Making predictions using model: %s", dlc.model)
    if len(psm_list_cal) > 0:
        preds = dlc.make_preds(seq_df=df_pred, infile=file_pred, psm_list=psm_list_pred)
    else:
        preds = dlc.make_preds(seq_df=df_pred, infile=file_pred, psm_list=psm_list_pred, calibrate=False)
    
    #df_pred["predicted_tr"] = preds
    logger.info("Writing predictions to file: %s", file_pred_out)
    
    file_pred_out = open(file_pred_out,"w")
    file_pred_out.write("Sequence proforma,predicted retention time\n")
    for psm,tr in zip(psm_list_pred,preds):
        file_pred_out.write(f"{psm.peptidoform.proforma},{tr}\n")
    file_pred_out.close()

    logger.info("DeepLC finished!")


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()


setup(
    name='deeplc',
    version='2.2.23',
    license='apache-2.0',
    description='DeepLC: Retention time prediction for (modified) peptides using Deep Learning.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Robbin Bouwmeester, Niels Hulstaert, Arthur Declercq, Ralf Gabriels, Prof. Lennart Martens, Prof. Sven Degroeve',
    author_email='Robbin.Bouwmeester@UGent.be',
    url='http://compomics.github.io/projects/DeepLC',
    project_urls={
        'Documentation': 'http://compomics.github.io/projects/DeepLC',
        'Source': 'https://github.com/compomics/DeepLC',
        'Tracker': 'https://github.com/compomics/DeepLC/issues'
    },
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'deeplc=deeplc.__main__:main',
            'deeplc-gui=deeplc.gui:start_gui',
        ]
    },
    keywords=[
        'DeepLC', 'Proteomics', 'deep learning', 'peptides', 'retention time',
        'prediction'
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 4 - Beta"
    ],
    install_requires=[
        'setuptools>=42.0.1',
        'tensorflow>=2.2,<2.13.0',
        'scipy>=1.4.1,<2',
        'numpy>=1.17,<2',
        'pandas>=0.25,<2',
        'matplotlib>=3,<4',
        'h5py>=2.10.0,<4',
        'pygam>=0.8.0,<1',
        'scikit-learn>=0.24.0,<2',
        'deeplcretrainer>=0.1,<1',
        'psm_utils>=0.2.3,<1',
        'hdf5plugin>=4.1.1'
    ],
    extras_require={
        "gui": ["gooey>=1.0"],
        "plot": ["plotly>=5"],
    },
    python_requires='>=3.7',
)
